from csv import reader
from typing import Dict, List

import numpy as np


def csv_line_to_dict(line: str, fieldnames: List, delimiter: str) -> Dict:
    return dict(
        zip(
            fieldnames,
            list(reader([line.decode("utf-8")], delimiter=delimiter))[0],
        )
    )


def csv_line_to_list(line: str, delimiter: str) -> List:
    return list(reader([line.decode("utf-8")], delimiter=delimiter))[0]


def index(
    path: str,
    delimiter: str = ",",
    fieldnames: List = None,
    has_header: bool = True,
    return_dict: bool = True,
    key_id: str = "id",
) -> Dict:
    assert (
        fieldnames or has_header
    ), "File must have header or fieldnames must be defined by user"

    index = {}

    with open(path, "rb") as file:
        position = file.tell()  # Init position

        for i, line in enumerate(file):
            if i == 0 and has_header:
                if not fieldnames:
                    fieldnames = line.decode("utf-8").strip().split(delimiter)

            elif return_dict:
                idx = csv_line_to_dict(
                    line=line, fieldnames=fieldnames, delimiter=delimiter
                )[key_id]
                index[idx] = position

            else:
                index[i] = position

            position = file.tell()  # Update position

    return fieldnames, index


def get(
    path: str,
    index: dict,
    idx: str,
    delimiter: str = ",",
    fieldnames: List = None,
    return_dict: bool = True,
) -> Dict:
    with open(path, "rb") as file:
        position = index[idx]
        file.seek(position)
        line = file.readline()

    return (
        csv_line_to_dict(line=line, fieldnames=fieldnames, delimiter=delimiter)
        if return_dict
        else csv_line_to_list(line=line, delimiter=delimiter)
    )


def mget(
    path: str,
    index: dict,
    indices: str,
    delimiter: str = ",",
    fieldnames: List = None,
    return_dict: bool = True,
) -> List[Dict]:
    positions = np.array([index[idx] for idx in indices])
    sorting_indices = np.argsort(positions)

    lines = [None] * len(positions)

    with open(path, "rb") as file:
        for i in sorting_indices:
            file.seek(positions[i])
            lines[i] = file.readline()

    if return_dict:
        return [
            csv_line_to_dict(
                line=line, fieldnames=fieldnames, delimiter=delimiter
            )
            for line in lines
        ]

    return [csv_line_to_list(line=line, delimiter=delimiter) for line in lines]



import os
from typing import Callable, Dict, List, Tuple, Union

import numpy as np
import orjson

from .handlers import (
    csv_handler,
    custom_handler,
    jsonl_handler,
    numpy_handler,
    txt_handler,
)


class Indxr:
    def __init__(
        self, path: str, kind="infer", callback: Callable = None, **kwargs: Dict
    ):
        # Init  ----------------------------------------------------------------
        self.kind = kind
        self.path = path
        self.kwargs = kwargs
        self.callback = callback
        self.index = None  # Dict : k -> file position
        self.index_keys = None  # List of index keys

        # Infer file extension -------------------------------------------------
        if self.kind == "infer":
            self.kind = os.path.splitext(self.path)[1][1:]

        if self.kind not in {"txt", "jsonl", "csv", "tsv", "custom", "dat"}:
            raise NotImplementedError(
                f"Specified `kind` not supported. {self.kind}"
            )

        # Init kwargs ----------------------------------------------------------
        if not self.kwargs:
            self.kwargs = {}

        if "delimiter" not in self.kwargs:
            self.kwargs["delimiter"] = "\t" if self.kind == "tsv" else ","

        if "fieldnames" not in self.kwargs:
            self.kwargs["fieldnames"] = None

        if "has_header" not in self.kwargs:
            self.kwargs["has_header"] = True

        if "return_dict" not in self.kwargs:
            self.kwargs["return_dict"] = True

        if "key_id" not in self.kwargs:
            self.kwargs["key_id"] = "id"

        # Create index ---------------------------------------------------------
        self.index = self.create_index()
        self.index_keys = list(self.index)

    def create_index(self) -> Dict:
        if self.kind == "txt":
            return txt_handler.index(self.path)

        elif self.kind == "jsonl":
            return jsonl_handler.index(self.path, self.kwargs["key_id"])

        elif self.kind in {"csv", "tsv"}:
            fieldnames, index = csv_handler.index(self.path, **self.kwargs)
            self.kwargs["fieldnames"] = fieldnames
            return index

        elif self.kind == "dat":
            return numpy_handler.index(
                self.kwargs["dtype"], self.kwargs["shape"]
            )

        elif self.kind == "custom":
            return custom_handler.index(self.path)

        raise NotImplementedError("Specified `kind` not supported.")

    def get(self, idx: Union[str, int]) -> Union[str, Dict, np.ndarray]:
        if self.kind == "txt":
            x = txt_handler.get(path=self.path, index=self.index, idx=idx)

        elif self.kind == "jsonl":
            x = jsonl_handler.get(path=self.path, index=self.index, idx=idx)

        elif self.kind in {"csv", "tsv"}:
            x = csv_handler.get(
                path=self.path,
                index=self.index,
                idx=idx,
                delimiter=self.kwargs["delimiter"],
                fieldnames=self.kwargs["fieldnames"],
                return_dict=self.kwargs["return_dict"],
            )

        elif self.kind == "dat":
            x = numpy_handler.get(
                path=self.path, index=self.index, idx=str(idx)
            )

        elif self.kind == "custom":
            x = custom_handler.get(path=self.path, index=self.index, idx=idx)

        else:
            raise NotImplementedError()

        return self.callback(x) if self.callback else x

    def mget(self, indices: List[str]) -> List[Union[str, Dict, np.ndarray]]:
        if self.kind == "txt":
            xs = txt_handler.mget(
                path=self.path, index=self.index, indices=indices
            )

        elif self.kind == "jsonl":
            xs = jsonl_handler.mget(
                path=self.path, index=self.index, indices=indices
            )

        elif self.kind in {"csv", "tsv"}:
            xs = csv_handler.mget(
                path=self.path,
                index=self.index,
                indices=indices,
                delimiter=self.kwargs["delimiter"],
                fieldnames=self.kwargs["fieldnames"],
                return_dict=self.kwargs["return_dict"],
            )

        elif self.kind == "dat":
            xs = numpy_handler.mget(
                path=self.path,
                index=self.index,
                indices=[str(idx) for idx in indices],
            )

        elif self.kind == "custom":
            xs = custom_handler.mget(
                path=self.path, index=self.index, indices=indices
            )

        else:
            raise NotImplementedError()

        return [self.callback(x) for x in xs] if self.callback else xs

    def get_slice(self, start: int, stop: int) -> np.ndarray:
        if self.kind == "dat":
            return numpy_handler.get_slice(
                path=self.path, index=self.index, start=start, stop=stop
            )

        else:
            raise NotImplementedError()

    def mget_slice(self, slices: List[Tuple[int]]) -> List[np.ndarray]:
        if self.kind == "dat":
            return numpy_handler.mget_slice(
                path=self.path, index=self.index, slices=slices
            )

        else:
            raise NotImplementedError()

    def write(self, path: str):
        with open(path, "wb") as f:
            f.write(
                orjson.dumps(
                    {
                        "path": self.path,
                        "kind": self.kind,
                        "kwargs": self.kwargs,
                        "index": self.index,
                        "index_keys": self.index_keys,
                    },
                    option=orjson.OPT_INDENT_2,
                )
            )

    @staticmethod
    def read(path: str, callback: Callable = None):
        with open(path, "rb") as f:
            x = orjson.loads(f.read())

        if "shape" in x["kwargs"]:
            x["kwargs"]["shape"] = tuple(x["kwargs"]["shape"])

        indxr = Indxr(path=x["path"], kind=x["kind"], **x["kwargs"])
        indxr.index = x["index"]
        indxr.index_keys = x["index_keys"]
        indxr.callback = callback

        return indxr

    def __getitem__(self, idx: int) -> Union[str, Dict]:
        return self.get(self.index_keys[idx])

    def __len__(self) -> int:
        return len(self.index_keys)



from typing import Dict, List

import numpy as np
import orjson


def index(path: str, key_id: str = "id") -> Dict:
    index = {}  # Init index dictionary

    with open(path, "rb") as file:
        position = file.tell()  # Init position

        for line in file:
            idx = orjson.loads(line)[key_id]
            index[idx] = position
            position = file.tell()  # Update position

    return index


def get(path: str, index: Dict, idx: str) -> Dict:
    with open(path, "rb") as file:
        position = index[idx]
        file.seek(position)
        line = file.readline()

    return orjson.loads(line)  # Convert line to Python Dictionary


def mget(path: str, index: Dict, indices: str) -> List[Dict]:
    positions = np.array([index[idx] for idx in indices])
    sorting_indices = np.argsort(positions)

    lines = [None] * len(positions)

    with open(path, "rb") as file:
        for i in sorting_indices:
            file.seek(positions[i])
            lines[i] = file.readline()

    return [orjson.loads(line) for line in lines]



from typing import Dict, List

import numpy as np


def index(path: str) -> Dict:
    index = {}  # Init index dictionary

    with open(path, "rb") as file:
        position = file.tell()  # Init position

        for i, _ in enumerate(file):
            index[str(i)] = position
            position = file.tell()  # Update position

    return index


def get(path: str, index: Dict, idx: int) -> bytes:
    print("ok")
    with open(path, "rb") as file:
        position = index[idx]
        file.seek(position)
        line = file.readline()

    return line.strip()


def mget(path: str, index: Dict, indices: str) -> List[bytes]:
    positions = np.array([index[idx] for idx in indices])
    sorting_indices = np.argsort(positions)

    lines = [None] * len(positions)

    with open(path, "rb") as file:
        for i in sorting_indices:
            file.seek(positions[i])
            lines[i] = file.readline()

    return [line.strip() for line in lines]



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="indxr",
    version="0.1.2",
    author="Elias Bassani",
    author_email="elias.bssn@gmail.com",
    description="indxr: A Python utility for indexing long files.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/AmenRa/indxr",
    packages=setuptools.find_packages(),
    install_requires=[
        "numpy",
        "orjson",
    ],
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Intended Audience :: Science/Research",
        "Operating System :: OS Independent",
        "Topic :: Text Processing :: General",
    ],
    keywords=[
        "text index",
        "file index",
        "index",
        "indexer",
        "indexing",
        "information retrieval",
        "natural language processing",
    ],
    python_requires=">=3.7",
)



import re
from typing import Dict, List, Tuple

import numpy as np


def get_itemsize(dtype: str):
    return int(re.sub("\D", "", dtype)) // 8


def index(dtype: str, shape: Tuple[int]) -> Dict:
    step = shape[1] * get_itemsize(dtype)
    index = {str(i): i * step for i in range(shape[0])}
    index["dim"] = shape[1]
    index["dtype"] = dtype
    return index


def get(path: str, index: Dict, idx: str) -> np.ndarray:
    return np.memmap(
        path,
        dtype=index["dtype"],
        mode="r",
        shape=(1, index["dim"]),
        offset=index[idx],
    )[0]


def mget(path: str, index: Dict, indices: List[str]) -> np.ndarray:
    return np.asarray([get(path, index, idx) for idx in indices])


def get_slice(path: str, index: Dict, start: int, stop: int) -> np.ndarray:
    return np.memmap(
        path,
        dtype=index["dtype"],
        mode="r",
        shape=(stop - start, index["dim"]),
        offset=index[str(start)],
    )


def mget_slice(path: str, index: Dict, slices: List[Tuple[int]]) -> np.ndarray:
    return [get_slice(path, index, *slice) for slice in slices]
